#include "all.h"
#include "ext-reg-comp.h"

extern unsigned int GGGgetopt_store;
unsigned char *GGGerrno_str_ptr;

#ifndef DIET //Not dietlibc.
extern struct obstack the_global_obstack; //The obstack.
extern struct obstack *GGGthe_global_obstack; //The global ptr.
#endif

#define INSERTSTMT "insert"

#ifdef PG_ERR //pgsql has a lib problem.
#define VARCHAROID      1043
#endif

extern const unsigned char *my_name;

//FIXME use at exit or something like that to clean up using this function.
void pg_cleanup(PGconn *pgconn, const unsigned char * const command)
{
    ExecStatusType com_status;
    PGresult *result;
    
    //Simple clean up function prior to exit or signaling.
    
    result = PQexec(pgconn, command);
    com_status = PQresultStatus(result);
    
    if(com_status != PGRES_COMMAND_OK)
    {
        GGGerrno_str_ptr = PQerrorMessage(pgconn);
        fprintf(stderr, _(&FPCDT), my_name, command);
        fprintf(stderr, _(&PQEMR), GGGerrno_str_ptr);
    }
}

unsigned long long int strcnul(const unsigned char * const str, \
                                                const size_t len )
{
    unsigned long long int c;
    unsigned long long int i;
    
    /*
     * Returns the amount of nuls in the array str.
     */
    
    for(i = 0, c = 0; i < len; i++)
    {
        if(str[i] == '\0')
        {
            c++;
        }
    }
    
    return(c);
}

void build_exts(const unsigned char *mime_types_file_name, PGconn *pgconn)
{
    static const signed int open_read = O_RDONLY;
    static const Oid paramtypes[2] = { VARCHAROID, VARCHAROID };
    static const int *paramformats[2] = { 0, 0 };
    static const char *insert_str = \
           "INSERT INTO mime_types " \
        "(mime, ext) VALUES ($1, $2);";
    unsigned long long int count;
    signed int mime_types_file;
    unsigned char *argz_str = NULL;
    const size_t size2read = 1;
    unsigned char string_to_be_database[256];
    unsigned char index = 0;
    ssize_t bytes_read;
    static size_t argz_len;
    static PGresult *result;
    static ExecStatusType com_status;
    bool ext = false;
    bool eof = false;
    bool eol = false;
    
    
    /*
     * First it sanity checks pgconn they it starts reading lines
     * discarding anything with a leading # in it. It then enters all the
     * data into a database. It has to use malloc to split the strings.
     */
    
    
    if(mime_types_file_name == NULL || pgconn == NULL)
    {
        //This should never happen.
        fprintf(stderr, _(&WRONG), my_name, __FILE__, __LINE__);
        (void)raise(SIGABRT);
    }
    
    //Check if the connection is ok.
    if(PQstatus(pgconn) == CONNECTION_BAD)
    {
        //This should never happen.
        fprintf(stderr, _(&WRONG), my_name, __FILE__, __LINE__);
        (void)raise(SIGABRT);
    }
    
    //Create the new table.
    result = PQexec(pgconn,     "CREATE TABLE mime_types ( "  \
    "mime varchar(255) NOT NULL, ext varchar(255) NOT NULL );" );
    
    com_status = PQresultStatus(result);
    
    if(com_status != PGRES_COMMAND_OK)
    {
        GGGerrno_str_ptr = PQerrorMessage(pgconn);
        fprintf(stderr, _(&FPCCT), my_name, &"CREATE TABLE mime_types ( "  \
                 "mime varchar(255) NOT NULL, ext varchar(255) NOT NULL );" );
        fprintf(stderr, _(&PQEMR), GGGerrno_str_ptr);
        
        pg_cleanup(pgconn, "DROP TABLE mime_types;");
        
        (void)raise(SIGABRT);
    }
    
    result = PQprepare(pgconn, INSERTSTMT, insert_str, 2, &paramtypes);
    
    com_status = PQresultStatus(result);
    
    if(com_status != PGRES_COMMAND_OK)
    {
        GGGerrno_str_ptr = PQerrorMessage(pgconn);
        fprintf(stderr, _(&FPCPS), my_name, insert_str );
        fprintf(stderr, _(&PQEMR), GGGerrno_str_ptr);
        
        pg_cleanup(pgconn, "DROP TABLE mime_types;");
        
        (void)raise(SIGABRT);
    }
    
    
    //Open the file checking for errors.
    do
    {
        errno = 0;
        mime_types_file = open(mime_types_file_name, open_read);
        
        if(mime_types_file == -1)
        {
            check_errnof( _(&FFPD), mime_types_file_name);
        }
    } while(errno == EINTR /*No EAGAIN for an open call*/);
    
    
    do
    {
        if(index == 255)
        {
            string_to_be_database[255] = '\0';
            fprintf(stderr, _(&ESTL), my_name, \
                mime_types_file_name, string_to_be_database);
        }
        
        errno = 0;
        bytes_read = read(mime_types_file, \
          &string_to_be_database[index], size2read);
        
        switch(bytes_read)
        {
            case 0:
                //EOF
                eof = true;
                continue;
            case -1:
                check_errnof( _(&FCRFF), mime_types_file_name);
            default:
                /*
                 * Do you get the funny feeling that control
                 * is supposed to go here?
                 */
                
                if(GGGgetopt_store & 0x3)
                {
                    //Debug.
                    fprintf(stderr, "%c", string_to_be_database[index]);
                }
                
                if(string_to_be_database[0] == '#')
                {
                    //A comment, continue.
                    
                    if(isspace(string_to_be_database[index]) && \
                           !isblank(string_to_be_database[index]))
                    {
                        if(eol)
                        {
                            /*
                             * This is the second line terminator
                             * (Hello Windowz.) Ignore it.
                             */
                            
                            eol = false;
                            index = 0;
                            //It needs to be something other then a '#'.
                            string_to_be_database[0] = '.';
                        }
                        else
                        {
                            //This is the first line terminator.
                            //Set eol to true.
                            eol = true;
                            index++;
                        }
                        
                        continue;
                    }
                    else
                    {
                        if(eol)
                        {
                            //This is the next line and it needs the first char.
                            string_to_be_database[0] = string_to_be_database[index];
                            index=1;
                            eol = false;
                            continue;
                        }
                        
                        index++;
                        continue;
                    }
                }
                
                //Check that the data is clean. All fall throughs are purposeful.
                switch(string_to_be_database[index])
                {
                    case '!':
                    case '`':
                    case '~':
                    case '@':
                    case '$':
                    case '%':
                    case '^':
                    case '&':
                    case '*':
                    case '(':
                    case ')':
                    case '{':
                    case '}':
                    case '[':
                    case ']':
                    case ':':
                    case ';':
                    case '"':
                    case '\'':
                    case '<':
                    case '>':
                    case ',':
                    case '?':
                    case '|':
                    case '\\':
                    //All the begining control chars except the newline ones.
                    case '\000':
                    case '\001':
                    case '\002':
                    case '\003':
                    case '\004':
                    case '\005':
                    case '\006':
                    case '\007':
                    case '\010':
                    case '\016':
                    case '\017':
                    case '\020':
                    case '\021':
                    case '\022':
                    case '\023':
                    case '\024':
                    case '\025':
                    case '\026':
                    case '\027':
                    case '\030':
                    case '\031':
                    case '\032':
                    case '\033':
                    case '\034':
                    case '\035':
                    case '\036':
                    case '\037':
                        fprintf(stderr, _(&ECNAF), my_name, \
                               string_to_be_database[index], \
                                         mime_types_file_name );
                        exit(EXIT_FAILURE);
                    default:
                        //Now check if the char is too big.
                        
                        if(string_to_be_database[index] > 126)
                        {
                            //It's too big.
                            fprintf(stderr, _(&ECNAF), my_name, \
                                   string_to_be_database[index], \
                                             mime_types_file_name );
                            exit(EXIT_FAILURE);
                        }
                }
                
                if(isspace(string_to_be_database[index]))
                {
                    string_to_be_database[index] = '\0';
                    
                    if(index > 0)
                    {
                        /*
                         * Finished getting the string enter
                         * it into the database.
                         */
                        
                        index = 0;
                        errno = 0;
                        errno = 0;
                        
                        //Split the string into two. Malloc.
                        errno = argz_create_sep(string_to_be_database, \
                                              '=', &argz_str, &argz_len );
                        
                        if(errno)
                        {
                            //This should never happen.
                            GGGerrno_str_ptr = strerror(errno);
                            fprintf(stderr, _(&ERRRE), my_name, \
                                                GGGerrno_str_ptr );
                            (void)raise(SIGABRT);
                        }
                        
                        //There should be no more or less then two strings.
                        count = strcnul(argz_str, argz_len);
                        
                        if(2 != count)
                        {
                            fprintf(stderr, _(&ESS), my_name, \
                                string_to_be_database, argz_len);
                            
                            exit(EXIT_FAILURE);
                        }
                        
                        result = PQexecPrepared(pgconn, INSERTSTMT, 2, \
                                        argz_str, NULL, paramformats, 0 );
                        
                        com_status = PQresultStatus(result);
                        
                        if(com_status != PGRES_COMMAND_OK)
                        {
                            GGGerrno_str_ptr = PQerrorMessage(pgconn);
                            fprintf(stderr, _(&FPCID), my_name);
                            fprintf(stderr, _(&PQEMR), GGGerrno_str_ptr);
                            
                            pg_cleanup(pgconn, "DROP TABLE mime_types;");
                            
                            (void)raise(SIGABRT);
                        }
                        
                        free(argz_str);
                        
                        continue;
                    }
                    else
                    {
                       //Blank line here.
                       continue; 
                    }
                }
                else
                {
                    //Retrieving mime continue.
                    index++;
                    continue;
                }
        }
        
        fprintf(stderr, my_name, _(&WRONG), __FILE__, __LINE__);
        (void)raise(SIGABRT);
        
    } while(!eof);
    
    
}


